/*******************************************************************************

 WINBINDER - A native Windows binding

 Copyright  2004 Hypervisual - see LICENSE.TXT for details
 Author: Rubem Pechansky (pechansky@hypervisual.com)

 Window creation and manipulation functions

*******************************************************************************/

//----------------------------------------------------------------- DEPENDENCIES

#include "wb.h"
#include <string.h>				// For stricmp()
#include <shellapi.h>			// For Shell_NotifyIcon()

//-------------------------------------------------------------------- CONSTANTS

#define MAXWINNAME				256
#define DEFAULT_WIN_STYLE		(WS_POPUP|WS_MINIMIZEBOX|WS_VISIBLE|WS_CLIPSIBLINGS|WS_CLIPCHILDREN|WS_MAXIMIZEBOX|WS_CAPTION|WS_BORDER|WS_SYSMENU|WS_THICKFRAME)

#define MODELESS_WINDOW_CLASS	"wbModelessDlg"
#define MODAL_WINDOW_CLASS		"wbModalDlg"
#define MAIN_WINDOW_CLASS		"wbMainWnd"

//----------------------------------------------------------------------- MACROS

#define CALL_CALLBACK(id,lp)	/* Call user function */ \
	{if(pwbobj && pwbobj->parent->pszCallBackFn && *pwbobj->parent->pszCallBackFn) { \
		wbCallUserFunction(pwbobj->parent->pszCallBackFn, pwbobj->parent, pwbobj, id, lp); }}

//----------------------------------------------------------- PRIVATE PROTOTYPES

static BOOL CenterWindow(HWND hwndDlg, HWND hwndParent);
static HICON GetWindowIcon(HWND hwnd);
static BOOL AddTaskBarIcon(HWND hwnd);
static BOOL DeleteTaskBarIcon(HWND hwnd);
static void UpdateLVlParams(HWND hwnd);
static int CALLBACK CompareLVItemsAscending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);
static int CALLBACK CompareLVItemsDescending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);

// Window procedures

static LRESULT CALLBACK DefaultWBProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK ModalWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK TabPageProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

//-------------------------------------------------------------------- VARIABLES

// Global

HINSTANCE hAppInstance = NULL;
BOOL bMinimizeToTaskBar = FALSE;

// Static

static int idStatusBar = 0;
static HWND hToolBar = NULL;
static HWND hwndListView = NULL;

//------------------------------------------------------------- PUBLIC FUNCTIONS

PWBOBJ wbCreateWindow(PWBOBJ pwboParent, UINT uWinBinderClass, LPCTSTR pszCaption, int xPos, int yPos, int nWidth, int nHeight, DWORD dwWinBinderStyle, long lParam)
{
	char szWindowName[MAXWINNAME] = "";
	LPTSTR pszClass;
	DWORD dwStyle = 0;
	DWORD dwExStyle = 0;
	PWBOBJ pwbo;

	// Initialize WinBinder object

	pwbo = wbAlloc(sizeof(WBOBJ));
	if(!pwbo)
		return NULL;

	pwbo->handle = NULL;
	pwbo->id = 0;
	pwbo->uClass = uWinBinderClass;
	pwbo->item = -1;
	pwbo->subitem = -1;
	pwbo->style = dwWinBinderStyle;
	pwbo->parent = pwboParent;
	pwbo->pszCallBackFn = NULL;
	pwbo->lparam = lParam;

	switch(uWinBinderClass) {

		case AppWindow:				// Fixed size main window
			pszClass = MAIN_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN | WS_MINIMIZEBOX;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_APPWINDOW;
			bMinimizeToTaskBar = (dwWinBinderStyle & WBC_COLLAPSED) != 0;
			break;

		case ResizableWindow:		// Resizable main window
			pszClass = MAIN_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPEDWINDOW | WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN | WS_SIZEBOX;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_APPWINDOW;
			bMinimizeToTaskBar = (dwWinBinderStyle & WBC_COLLAPSED) != 0;
			break;

		case PopupWindow:			// Fixed size main window, no minimize button
			pszClass = MAIN_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_APPWINDOW;
			bMinimizeToTaskBar = (dwWinBinderStyle & WBC_COLLAPSED) != 0;
			break;

		case ModalDialog:			// Modal dialog box
			pszClass = MODAL_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_POPUP | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN | DS_MODALFRAME;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_DLGMODALFRAME;
			break;

		case ModelessDialog:		// Modeless dialog box
			pszClass = MODELESS_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPEDWINDOW | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN;
//			WS_POPUP | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN;
			dwExStyle = dwExStyle ? dwExStyle : 0;
			break;

		case ToolDialog:			// Modeless tool dialog
			pszClass = MODELESS_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_POPUP | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_PALETTEWINDOW;
			break;

		default:
			return NULL;
	}

	// Create the window

	if(BITTEST(dwWinBinderStyle, WBC_INVISIBLE))
		dwStyle = dwStyle & (~WS_VISIBLE);

	strncpy(szWindowName, pszCaption, MAXWINNAME - 1);
	pwbo->handle = CreateWindowEx(
		dwExStyle,
		pszClass,
		szWindowName,
		dwStyle ? dwStyle : DEFAULT_WIN_STYLE,
		xPos < 0 ? CW_USEDEFAULT : xPos,
		yPos < 0 ? CW_USEDEFAULT : yPos,
		nWidth <= 0 ? CW_USEDEFAULT : nWidth,
		nHeight <= 0 ? CW_USEDEFAULT : nHeight,
		pwboParent ? pwboParent->handle : NULL,
		NULL, hAppInstance, NULL);

	if(!pwbo->handle) {
		wbError(MB_ICONWARNING, "Error creating window");
		return NULL;
	}

	SetWindowLong(pwbo->handle, GWL_USERDATA, (LONG)pwbo);

	// Is it a modal dialog?

	if(uWinBinderClass == ModalDialog)
		SendMessage(pwbo->handle, WM_SETFOCUS, (WPARAM)GetWindow(pwbo->handle, GW_CHILD), 0);

	return pwbo;
}

BOOL wbSetWindowHandler(PWBOBJ pwbo, LPCTSTR pszHandler)
{
	HWND hChild = NULL;
	PWBOBJ poChild;

	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	if(!pszHandler || !*pszHandler)
		return FALSE;

	pwbo->pszCallBackFn = (LPTSTR)pszHandler;

	// Look for child tab controls to assign the same handler to them

	while((hChild = FindWindowEx(pwbo->handle, hChild, WC_TABCONTROL, NULL)) != NULL) {
		poChild = wbGetWBObj(hChild);
		poChild->pszCallBackFn = poChild->parent->pszCallBackFn;
	}

	return TRUE;
}

BOOL wbDestroyWindow(PWBOBJ pwbo)
{
	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	return DestroyWindow(pwbo->handle);
}

BOOL wbSetWindowSize(PWBOBJ pwbo, int nWidth, int nHeight)
{
	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	nWidth = MAX(0, nWidth);
	nHeight = MAX(0, nHeight);
	return SetWindowPos(pwbo->handle, 0, 0, 0, nWidth, nHeight, SWP_NOMOVE | SWP_NOZORDER);
}

/*
	TODO: use hwndParent
	TODO: Center child windows on their parents
	TODO: Center horizontally only or vertically only according to xPos, yPos
*/

BOOL wbSetWindowPosition(PWBOBJ pwbo, int xPos, int yPos, PWBOBJ pwboParent)
{
	if(xPos == WBC_CENTER || yPos == WBC_CENTER)
		return CenterWindow(pwbo->handle, pwboParent->handle);
	else
		return SetWindowPos(pwbo->handle, 0, xPos, yPos, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

/* Sort a list view column */

BOOL wbSortLVColumn(PWBOBJ pwbo, int nSubItem, BOOL bAscending)
{
	hwndListView = pwbo->handle;		// For CompareLVItems()

	SendMessage(pwbo->handle, LVM_SORTITEMS, nSubItem, (bAscending ?
	    (LPARAM)(PFNLVCOMPARE)CompareLVItemsAscending :
	    (LPARAM)(PFNLVCOMPARE)CompareLVItemsDescending));
	UpdateLVlParams(hwndListView);
	return TRUE;
}

/* Make uPeriod equal to zero to kill the timer */

BOOL wbSetTimer(PWBOBJ pwbo, int id, UINT uPeriod)
{
	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	if(!uPeriod)
		return KillTimer(pwbo->handle, id);

	if(SetTimer(pwbo->handle, id, uPeriod, NULL))
		return TRUE;
	else
		return FALSE;
}

//------------------------------------------- FUNCTIONS PUBLIC TO WINBINDER ONLY

void SetToolBarHandle(HWND hCtrl)
{
	hToolBar = hCtrl;
}

void SetStatusBarId(int id)
{
	idStatusBar = id;
}

/* Register a window class */

BOOL RegisterClasses(void)
{
	WNDCLASS wc;

	// Main window class

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = /*CS_HREDRAW | CS_VREDRAW | */CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)MainWndProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.lpszClassName = MAIN_WINDOW_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);	//IDI_WINLOGO
	wc.cbWndExtra = 0;

	if(!RegisterClass(&wc))
		return FALSE;

	// Modal dialogs

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = /*CS_HREDRAW | CS_VREDRAW | */CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)ModalWndProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.lpszClassName = MODAL_WINDOW_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = NULL;//LoadIcon(NULL, IDI_WINLOGO);
	wc.cbWndExtra = DLGWINDOWEXTRA;

	if(!RegisterClass(&wc))
		return FALSE;

	// Modeless dialogs

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = /*CS_HREDRAW | CS_VREDRAW | */CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)ModalWndProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.lpszClassName = MODELESS_WINDOW_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = NULL;//LoadIcon(NULL, IDI_WINLOGO);
	wc.cbWndExtra = DLGWINDOWEXTRA;

	if(!RegisterClass(&wc))
		return FALSE;

	// Tab pages

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)TabPageProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);//GetStockObject(HOLLOW_BRUSH);
	wc.lpszClassName = TAB_PAGE_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = NULL;//LoadIcon(NULL, IDI_WINLOGO);
	wc.cbWndExtra = DLGWINDOWEXTRA;

	if(!RegisterClass(&wc))
		return FALSE;

	return TRUE;
}

//------------------------------------------------------------ PRIVATE FUNCTIONS

/* TODO: hwndDlg is now centered on screen, hwndParent is ignored */

static BOOL CenterWindow(HWND hwndDlg, HWND hwndParent)
{
	HWND hwndOwner;
	RECT rcDlg, rcOwner, rc;

	// From Windows Help, "Initializing a Dialog Box"

	// Get the owner window and dialog box rectangles.

	if((hwndOwner = GetParent(hwndDlg)) == NULL)
		hwndOwner = GetDesktopWindow();
	GetWindowRect(hwndOwner, &rcOwner);
	GetWindowRect(hwndDlg, &rcDlg);
	CopyRect(&rc, &rcOwner);

	// Offset the owner and dialog box rectangles so that
	// right and bottom values represent the width and
	// height, and then offset the owner again to discard
	// space taken up by the dialog box.

	OffsetRect(&rcDlg, -rcDlg.left, -rcDlg.top);
	OffsetRect(&rc, -rc.left, -rc.top);
	OffsetRect(&rc, -rcDlg.right, -rcDlg.bottom);

	// The new position is the sum of half the remaining
	// space and the owner's original position.

	SetWindowPos(hwndDlg,
		HWND_TOP,
		rcOwner.left + (rc.right / 2),
		rcOwner.top + (rc.bottom / 2),
		0, 0,          // ignore size arguments
		SWP_NOSIZE);

	return TRUE;
}

/*
	The all-important common processing routine for all parent windows

	TODO: Filter user commands. Try: (LPNMHDR)lParam)->code == NM_CLICK, etc.
	TODO: Use the return value as a flag to know whether the calling function should continue default processing or not
	TODO: Use pwbobj->uClass instead of GetClassName()

*/

static LRESULT CALLBACK DefaultWBProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HWND hTBWnd = NULL;			// Handle of toolbar window

	switch(msg) {

		case WM_NOTIFY:
			{
				char szClass[256];
				HWND hCtrl = ((LPNMHDR)lParam)->hwndFrom;
				PWBOBJ pwbobj;

				if(!IsWindow(hCtrl))
					break;

				// printf("[%d] ", ((LPNMHDR)lParam)->code);

				// The classes below don't call the callback function

				GetClassName(hCtrl, szClass, 256);

				if(!stricmp(szClass, TOOLBARCLASSNAME)) {			// Toolbar

					if(!hTBWnd)
						hTBWnd = ((LPNMHDR)lParam)->hwndFrom;
					break;

				} else if(!stricmp(szClass, TOOLTIPS_CLASS)) {		// Tooltip

					if(((LPNMHDR)lParam)->code == (UINT)TTN_NEEDTEXT) {
						if(hTBWnd) {
							TBBUTTON tbb;
							int index;

							index = SendMessage(hTBWnd, TB_COMMANDTOINDEX,
							  ((LPTOOLTIPTEXT)lParam)->hdr.idFrom, 0);
							SendMessage(hTBWnd, TB_GETBUTTON, index, (LPARAM)(LPTBBUTTON)&tbb);
							((LPTOOLTIPTEXT)lParam)->lpszText = (LPTSTR)tbb.dwData;
						}
					}
					break;
				}

				// Verify if window is a WinBinder object

				pwbobj = wbGetWBObj(hCtrl);

				if(!pwbobj || !pwbobj->parent)
					break;

				if(!pwbobj->parent->pszCallBackFn)
					break;

				// Call callback function according to WinBinder class

				switch(pwbobj->uClass) {

					case Spinner:

						if(((LPNMHDR)lParam)->code == (UINT)UDN_DELTAPOS) {
							CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, 0);
						}
						break;

					case TreeView:

						switch(((LPNMHDR)lParam)->code) {

							case NM_DBLCLK:
								//printf("%d\n", ((LPNMHDR)lParam)->code);
								CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, WBC_DBLCLICK);
								break;

							case (UINT)TVN_SELCHANGED:
								CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, 0);
								break;
						}
						break;

					case TabControl:

						if(((LPNMHDR)lParam)->code == (UINT)TCN_SELCHANGE) {

							HWND hTab = ((LPNMHDR)lParam)->hwndFrom;
							int nSelTab = TabCtrl_GetCurSel(hTab);

							wbSelectTab(wbGetWBObj(hCtrl), nSelTab);
							CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, nSelTab);
							return 0;
						}
						break;

					case ListView:

						switch(((LPNMHDR)lParam)->code) {

							case NM_DBLCLK:
								CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, WBC_DBLCLICK);
								break;

							case LVN_ITEMCHANGED:

								if(((LPNM_LISTVIEW)lParam)->uChanged & (LVIF_STATE | LVIS_CHECKED)) {
									CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, 0);
								}
								break;

							case LVN_COLUMNCLICK:

								hwndListView = pwbobj->handle;		// For CompareLVItems()
								SendMessage(pwbobj->handle, LVM_SORTITEMS,
									((NM_LISTVIEW FAR *)lParam)->iSubItem, (LPARAM)(PFNLVCOMPARE)CompareLVItemsAscending);
								UpdateLVlParams(hwndListView);
								break;
						}
						break;

				} // switch(pwbobj->uClass)
			} // ~WM_NOTIFY
			break;

		case WM_HSCROLL:		// Scroll bars
		case WM_VSCROLL:

			{
				PWBOBJ pwbobj;
				HWND hCtrl;
				int nMin, nMax, nPos;

				hCtrl = (HWND)lParam;
				pwbobj = wbGetWBObj(hCtrl);

				if(!pwbobj)
					break;

				nPos = SendMessage(hCtrl, SBM_GETPOS, 0, 0);

				switch(LOWORD(wParam)) {
					case SB_LINEUP:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos -= (nMax - nMin) / 100;
						nPos = MAX(nPos, 0);
						break;

					case SB_LINEDOWN:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos += (nMax - nMin) / 100;
						nPos = MAX(nPos, 0);
						break;

					case SB_PAGEUP:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos -= (nMax - nMin) / 10;
						nPos = MAX(nPos, 0);
						break;

					case SB_PAGEDOWN:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos += (nMax - nMin) / 10;
						nPos = MAX(nPos, 0);
						break;

					case SB_THUMBPOSITION:
					case SB_THUMBTRACK:						// Same value as TB_THUMBTRACK
						nPos = (short)HIWORD(wParam);
						break;
				}

				SendMessage(hCtrl, SBM_SETPOS, nPos, TRUE);
				CALL_CALLBACK(pwbobj->id, 0);
			}
			break;

		case WM_COMMAND:
			{
				BOOL bProcess;
				HWND hCtrl;
				PWBOBJ pwbobj;

				if((HWND)lParam == 0 && (HIWORD(wParam) == 0 || HIWORD(wParam) == 1)) {		// Is the message from a menu or accelerator?

					pwbobj = wbGetWBObj(hwnd);

					if(pwbobj && pwbobj->pszCallBackFn && *pwbobj->pszCallBackFn)
						wbCallUserFunction(pwbobj->pszCallBackFn, pwbobj, pwbobj, LOWORD(wParam), 0);
					// Must not use the macro CALL_CALLBACK here
					break;
				}

				hCtrl = (HWND)lParam;
				pwbobj = wbGetWBObj(hCtrl);

				if(!pwbobj || !pwbobj->parent || !pwbobj->parent->pszCallBackFn)
					break;

				switch(pwbobj->uClass) {	// Classes that receive their notifications via WM_COMMAND

					case EditBox:
					case RTFEditBox:
						bProcess = (HIWORD(wParam) == EN_CHANGE);	// EN_UPDATE?
						CALL_CALLBACK(LOWORD(wParam), 0);
						break;

					case ComboBox:
						bProcess = (HIWORD(wParam) == CBN_SELCHANGE);
						CALL_CALLBACK(LOWORD(wParam), 0);
						break;

					case ListBox:
						if(HIWORD(wParam) == LBN_DBLCLK) {
							CALL_CALLBACK(LOWORD(wParam), WBC_DBLCLICK);
						} else {
							bProcess = (HIWORD(wParam) == LBN_SELCHANGE);
							CALL_CALLBACK(LOWORD(wParam), 0);
						}
						break;

					default:				// TODO: Maybe delete these lines,
						bProcess = TRUE;	// but must be sure first
						CALL_CALLBACK(LOWORD(wParam), 0);
				}
				return 0;
			}
			break;

		case WM_TIMER:

			{
				PWBOBJ pwbobj;

				pwbobj = wbGetWBObj(hwnd);

				if(!pwbobj || !pwbobj->pszCallBackFn)
					break;

				if(pwbobj->pszCallBackFn)
					wbCallUserFunction(pwbobj->pszCallBackFn, pwbobj, pwbobj, wParam, 0);
				return 0;
			}
			break;

	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}

static LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

		case WM_CREATE:
			if(((LPCREATESTRUCT)lParam)->x == WBC_CENTER && ((LPCREATESTRUCT)lParam)->x == WBC_CENTER)
				CenterWindow(hwnd, NULL);
			break;

		case WM_ACTIVATE:
			hCurrentDlg = hwnd;	// Used in IsDialogMessage() -- main loop
			break;

		case WM_SIZE:
			if(bMinimizeToTaskBar && (wParam == SIZE_MINIMIZED)) {
				AddTaskBarIcon(hwnd);
				ShowWindow(hwnd, SW_HIDE);
			}
			SendDlgItemMessage(hwnd, idStatusBar, msg, wParam, lParam);
			SendMessage(hToolBar, msg, wParam, lParam);
			break;

		case WM_CLOSE:
			{
				PWBOBJ pwbobj = (PWBOBJ)GetWindowLong(hwnd, GWL_USERDATA);

				if(pwbobj && pwbobj->pszCallBackFn && *pwbobj->pszCallBackFn) {

					// This message is only sent to windows that have callback functions

					SendMessage(hwnd, WM_COMMAND, MAKELONG(IDCLOSE, 0), (LPARAM)NULL);
					return 0;
				} else {
		 			break;
				}
 			}
			break;

		case WBWM_NOTIFYICON:
			if(bMinimizeToTaskBar) {

				switch (lParam){

					case WM_LBUTTONDBLCLK:
					case WM_LBUTTONUP:

						ShowWindow(hwnd, SW_SHOWMINNOACTIVE);
						ShowWindow(hwnd, SW_NORMAL);
						SetForegroundWindow(hwnd);
						SetFocus(hwnd);
						DeleteTaskBarIcon(hwnd);
						return TRUE;
				}
			}
			break;

		case WM_DESTROY:
			if(bMinimizeToTaskBar)
				DeleteTaskBarIcon(hwnd);
			PostQuitMessage(0);
			break;

		default:
			return DefaultWBProc(hwnd, msg, wParam, lParam);
	}
	return DefWindowProc(hwnd, msg, wParam, lParam);
}

// NOTE: USED FOR BOTH MODAL AND MODELESS WINDOWS
// The word "modal" below is fake: these are not "real" modal dialog boxes

static LRESULT CALLBACK ModalWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

		case WM_CREATE:
			if(((LPCREATESTRUCT)lParam)->x == WBC_CENTER && ((LPCREATESTRUCT)lParam)->x == WBC_CENTER)
				CenterWindow(hwnd, NULL);
			EnableWindow(GetParent(hwnd), FALSE);
			break;

		case WM_ACTIVATE:
			hCurrentDlg = hwnd;	// Used in IsDialogMessage() -- main loop
			break;

		case WM_CLOSE:
			{
				HWND hParent;

				hParent = GetParent(hwnd);
				if(hParent) {
					EnableWindow(hParent, TRUE);
					SetFocus(hParent);
				}
				DestroyWindow(hwnd);
				hwnd = NULL;
			}
			break;

		case WM_DESTROY:
			EnableWindow(GetParent(hwnd), TRUE);
			SetFocus(GetParent(hwnd));
			hCurrentDlg = NULL;
			break;

		default:
			return DefaultWBProc(hwnd, msg, wParam, lParam);
	}
	return DefDlgProc(hwnd, msg, wParam, lParam);
}

// Processes messages for tab pages

static LRESULT CALLBACK TabPageProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

//		case WM_CLOSE:		// So that an ESC does not close this page
//			return 0;

/*		case WM_COMMAND:
		printf("%d\n", LOWORD(wParam));
			if(LOWORD(wParam) == IDOK)
				return 0;
			break;*/

		case WM_SHOWWINDOW:
			if(wParam) {
				HWND hwndNext;

				hCurrentDlg = hwnd;	// Used in IsDialogMessage() -- main loop
				hwndNext = GetNextDlgTabItem(hwnd, NULL, FALSE);
				if(hwndNext)
					SetFocus(hwndNext);
			} else
				hCurrentDlg = NULL;
			break;
	}
	return DefaultWBProc(hwnd, msg, wParam, lParam);
}

// Adapted from from http://groups.google.com/groups?hl=en&lr=&selm=38BC4F60.11F62F%40thematic.com
// Peter Forbes / David Lowndes

// Try several methods to retrieve the icon from an application window

static HICON GetWindowIcon(HWND hwnd)
{
	HICON hIcon;

	hIcon = (HICON)SendMessage(hwnd, WM_GETICON, ICON_SMALL, 0);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)SendMessage(hwnd, WM_GETICON, ICON_BIG, 0);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)SendMessage(hwnd, WM_QUERYDRAGICON, 0, 0);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)GetClassLong(hwnd, GCL_HICONSM);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)GetClassLong(hwnd, GCL_HICON);
	if(hIcon)
		return hIcon;

	return LoadIcon(NULL, IDI_APPLICATION);
}

// From Win32 programmer's reference, "Adding and Deleting Taskbar Icons" and
// http://msdn.microsoft.com/library/en-us/shellcc/platform/shell/programmersguide/shell_int/shell_int_programming/taskbar.asp

static BOOL AddTaskBarIcon(HWND hwnd)
{
    BOOL bRes;
	NOTIFYICONDATA tnid;

	if(!hwnd)
		return FALSE;

	tnid.cbSize = sizeof(NOTIFYICONDATA);
	tnid.hWnd = hwnd;
	tnid.uID = 0;
	tnid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
	tnid.uCallbackMessage = WBWM_NOTIFYICON;
	tnid.hIcon = CopyIcon(GetWindowIcon(hwnd));
	GetWindowText(hwnd, tnid.szTip, 63);		// szTip[] is 64 bytes long
	bRes = Shell_NotifyIcon(NIM_ADD, &tnid);
	if(tnid.hIcon)
		DestroyIcon(tnid.hIcon);
	return bRes;
}

// From Win32 programmer's reference, "Adding and Deleting Taskbar Icons" and
// http://msdn.microsoft.com/library/en-us/shellcc/platform/shell/programmersguide/shell_int/shell_int_programming/taskbar.asp

static BOOL DeleteTaskBarIcon(HWND hwnd)
{
	BOOL bRes;
	NOTIFYICONDATA tnid;

	if(!hwnd)
		return FALSE;
	tnid.cbSize = sizeof(NOTIFYICONDATA);
	tnid.hWnd = hwnd;
	tnid.uID = 0;
	bRes = Shell_NotifyIcon(NIM_DELETE, &tnid);
	DestroyIcon(tnid.hIcon);
	return bRes;
}

/*

  Ported from Iczelion's Win32 Assembly Homepage: http://spiff.tripnet.se/~iczelion/tut31.html

  Quote: "Note that when the items are rearranged, their indexes also change. So when the sorting is done
  in my example, I need to update the values in lParam to reflect the new indexes."

*/

static void UpdateLVlParams(HWND hwnd)
{
	LV_ITEM lvi;
	int count;

	count = SendMessage(hwnd, LVM_GETITEMCOUNT, 0, 0);
	lvi.mask = LVIF_PARAM;
	lvi.iSubItem = 0;
	for(lvi.iItem = 0; lvi.iItem < count; lvi.iItem++) {
		lvi.lParam = lvi.iItem;
		SendMessage(hwnd, LVM_SETITEM, 0, (LPARAM)&lvi);
	}
}

/*

 This is the callback function that compares two ListView items.

 - lParam1 && lParam2 contain the item indices
 - lParamSort is the sub-item index
 - 64 characters is more than enough for string comparison

*/

static BOOL IsDecimalNumber(LPTSTR szStr)
{
	int i;
	int nLen = MIN(strlen(szStr), 63);

	for(i = 0; i < nLen; i++)
		if((*(szStr + i) < '0' || *(szStr + i) > '9') && *(szStr + i) != ' ')
			return FALSE;
	return TRUE;
}

static int CALLBACK CompareLVItemsAscending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	char szStr1[64];
	char szStr2[64];

	ListView_GetItemText(hwndListView, lParam1, lParamSort, szStr1, 63);
	ListView_GetItemText(hwndListView, lParam2, lParamSort, szStr2, 63);

	if(!*szStr1 && !*szStr2)				// Two empty strings
		return 0;
	else if(!*szStr2)						// Second string is empty
		return 1;
	else if(!*szStr1)						// First string is empty
		return -1;
	else if(IsDecimalNumber(szStr1) && IsDecimalNumber(szStr2))		// Two integers
		return(atol(szStr1) == atol(szStr2) ? 0 : ((atol(szStr1) > atol(szStr2)) ? 1 : -1));
	else
		return lstrcmp(szStr1, szStr2);		// According to locale, not case-sensitive
}

static int CALLBACK CompareLVItemsDescending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	char szStr1[64];
	char szStr2[64];

	ListView_GetItemText(hwndListView, lParam1, lParamSort, szStr1, 63);
	ListView_GetItemText(hwndListView, lParam2, lParamSort, szStr2, 63);

	if(!*szStr1 && !*szStr2)				// Two empty strings
		return 0;
	else if(!*szStr2)						// Second string is empty
		return -1;
	else if(!*szStr1)						// First string is empty
		return 1;
	else if(IsDecimalNumber(szStr1) && IsDecimalNumber(szStr2))		// Two integers
		return(atol(szStr1) == atol(szStr2) ? 0 : ((atol(szStr1) > atol(szStr2)) ? -1 : 1));
	else
		return lstrcmp(szStr2, szStr1);		// According to locale, not case-sensitive
}

//------------------------------------------------------------------ END OF FILE
